# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: api.proto
# plugin: python-betterproto
from dataclasses import dataclass
from datetime import datetime
from typing import AsyncGenerator, Dict, List, Optional

import betterproto
import grpclib


class MarketStatus(betterproto.Enum):
    MS_UNKNOWN = 0
    MS_ONLINE = 1


class Side(betterproto.Enum):
    S_UNKNOWN = 0
    S_BID = 1
    S_ASK = 2


class OrderType(betterproto.Enum):
    OT_MARKET = 0
    OT_LIMIT = 1
    OT_IOC = 2
    OT_POST = 3


class OrderStatus(betterproto.Enum):
    OS_UNKNOWN = 0
    OS_OPEN = 1
    OS_PARTIAL_FILL = 2
    OS_CANCELLED = 3
    OS_FILLED = 4


class Direction(betterproto.Enum):
    D_ASCENDING = 0
    D_DESCENDING = 1


class Step(betterproto.Enum):
    STEP0 = 0
    STEP1 = 1
    STEP2 = 2
    STEP3 = 3


@dataclass
class GetMarketsRequest(betterproto.Message):
    pass


@dataclass
class GetMarketsResponse(betterproto.Message):
    markets: Dict[str, "Market"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


@dataclass
class Market(betterproto.Message):
    market: str = betterproto.string_field(1)
    status: "MarketStatus" = betterproto.enum_field(2)
    address: str = betterproto.string_field(3)
    base_mint: str = betterproto.string_field(4)
    quoted_mint: str = betterproto.string_field(5)
    base_decimals: int = betterproto.int64_field(6)
    quote_decimals: int = betterproto.int64_field(7)


@dataclass
class GetTickersRequest(betterproto.Message):
    market: str = betterproto.string_field(1)


@dataclass
class GetTickersResponse(betterproto.Message):
    tickers: List["Ticker"] = betterproto.message_field(1)


@dataclass
class Ticker(betterproto.Message):
    market: str = betterproto.string_field(1)
    market_address: str = betterproto.string_field(2)
    bid: float = betterproto.double_field(3)
    bid_size: float = betterproto.double_field(4)
    ask: float = betterproto.double_field(5)
    ask_size: float = betterproto.double_field(6)


@dataclass
class GetKlineRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    from_: datetime = betterproto.message_field(2)
    to: datetime = betterproto.message_field(3)
    resolution: str = betterproto.string_field(4)
    limit: int = betterproto.uint32_field(5)


@dataclass
class GetKlineResponse(betterproto.Message):
    market: str = betterproto.string_field(1)
    timestamp: datetime = betterproto.message_field(2)
    candles: List["Candle"] = betterproto.message_field(3)


@dataclass
class Candle(betterproto.Message):
    start_time: datetime = betterproto.message_field(1)
    update_time: datetime = betterproto.message_field(2)
    open: float = betterproto.double_field(3)
    close: float = betterproto.double_field(4)
    low: float = betterproto.double_field(5)
    high: float = betterproto.double_field(6)
    amount: float = betterproto.double_field(7)
    volume: float = betterproto.double_field(8)
    count: float = betterproto.double_field(9)


@dataclass
class GetOrderbookRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    limit: int = betterproto.uint32_field(2)


@dataclass
class GetOrderbooksRequest(betterproto.Message):
    markets: List[str] = betterproto.string_field(1)
    limit: int = betterproto.uint32_field(2)


@dataclass
class GetOrderbookResponse(betterproto.Message):
    market: str = betterproto.string_field(1)
    market_address: str = betterproto.string_field(2)
    bids: List["OrderbookItem"] = betterproto.message_field(3)
    asks: List["OrderbookItem"] = betterproto.message_field(4)


@dataclass
class OrderbookItem(betterproto.Message):
    price: float = betterproto.double_field(1)
    size: float = betterproto.double_field(2)


@dataclass
class GetTradesRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    limit: int = betterproto.uint32_field(2)


@dataclass
class GetTradesResponse(betterproto.Message):
    trades: List["Trade"] = betterproto.message_field(1)


@dataclass
class Trade(betterproto.Message):
    side: "Side" = betterproto.enum_field(1)
    size: float = betterproto.double_field(2)
    price: float = betterproto.double_field(3)
    order_i_d: str = betterproto.string_field(4)
    is_maker: bool = betterproto.bool_field(5)


@dataclass
class GetServerTimeRequest(betterproto.Message):
    pass


@dataclass
class GetServerTimeResponse(betterproto.Message):
    timestamp: str = betterproto.string_field(1)


@dataclass
class GetAccountBalanceRequest(betterproto.Message):
    owner_address: str = betterproto.string_field(1)


@dataclass
class GetAccountBalanceResponse(betterproto.Message):
    tokens: List["TokenBalance"] = betterproto.message_field(1)


@dataclass
class TokenBalance(betterproto.Message):
    symbol: str = betterproto.string_field(1)
    address: str = betterproto.string_field(2)
    wallet_amount: float = betterproto.double_field(3)
    unsettled_amount: float = betterproto.double_field(4)
    open_orders_amount: float = betterproto.double_field(5)


@dataclass
class PostOrderRequest(betterproto.Message):
    owner_address: str = betterproto.string_field(1)
    payer_address: str = betterproto.string_field(2)
    market: str = betterproto.string_field(3)
    side: "Side" = betterproto.enum_field(4)
    type: List["OrderType"] = betterproto.enum_field(5)
    amount: float = betterproto.double_field(6)
    price: float = betterproto.double_field(7)
    open_orders_address: str = betterproto.string_field(8)
    client_order_i_d: int = betterproto.uint64_field(9)


@dataclass
class PostReplaceOrderRequest(betterproto.Message):
    owner_address: str = betterproto.string_field(1)
    payer_address: str = betterproto.string_field(2)
    market: str = betterproto.string_field(3)
    side: "Side" = betterproto.enum_field(4)
    type: List["OrderType"] = betterproto.enum_field(5)
    amount: float = betterproto.double_field(6)
    price: float = betterproto.double_field(7)
    open_orders_address: str = betterproto.string_field(8)
    client_order_i_d: int = betterproto.uint64_field(9)
    order_i_d: str = betterproto.string_field(10)


@dataclass
class PostOrderResponse(betterproto.Message):
    transaction: str = betterproto.string_field(1)
    open_orders_address: str = betterproto.string_field(2)


@dataclass
class PostCancelOrderRequest(betterproto.Message):
    order_i_d: str = betterproto.string_field(1)
    side: "Side" = betterproto.enum_field(2)
    market_address: str = betterproto.string_field(3)
    owner_address: str = betterproto.string_field(4)
    open_orders_address: str = betterproto.string_field(5)


@dataclass
class PostCancelByClientOrderIDRequest(betterproto.Message):
    client_order_i_d: int = betterproto.uint64_field(1)
    market_address: str = betterproto.string_field(2)
    owner_address: str = betterproto.string_field(3)
    open_orders_address: str = betterproto.string_field(4)


@dataclass
class PostCancelOrderResponse(betterproto.Message):
    transaction: str = betterproto.string_field(1)


@dataclass
class PostCancelAllRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    owner_address: str = betterproto.string_field(2)
    open_orders_addresses: List[str] = betterproto.string_field(3)


@dataclass
class PostCancelAllResponse(betterproto.Message):
    transactions: List[str] = betterproto.string_field(1)


@dataclass
class PostSettleRequest(betterproto.Message):
    owner_address: str = betterproto.string_field(1)
    market: str = betterproto.string_field(2)
    base_token_wallet: str = betterproto.string_field(3)
    quote_token_wallet: str = betterproto.string_field(4)
    open_orders_address: str = betterproto.string_field(5)


@dataclass
class PostSettleResponse(betterproto.Message):
    transaction: str = betterproto.string_field(1)


@dataclass
class Settlement(betterproto.Message):
    symbol: str = betterproto.string_field(1)
    unsettled: float = betterproto.double_field(2)
    amount: float = betterproto.double_field(3)


@dataclass
class GetOrdersRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    status: "OrderStatus" = betterproto.enum_field(2)
    side: "Side" = betterproto.enum_field(3)
    types: List["OrderType"] = betterproto.enum_field(4)
    from_: datetime = betterproto.message_field(5)
    limit: int = betterproto.uint32_field(6)
    direction: "Direction" = betterproto.enum_field(7)
    address: str = betterproto.string_field(8)
    open_orders_address: str = betterproto.string_field(9)


@dataclass
class GetOrdersResponse(betterproto.Message):
    orders: List["Order"] = betterproto.message_field(1)


@dataclass
class Order(betterproto.Message):
    order_i_d: str = betterproto.string_field(1)
    market: str = betterproto.string_field(2)
    side: "Side" = betterproto.enum_field(3)
    types: List["OrderType"] = betterproto.enum_field(4)
    price: float = betterproto.double_field(5)
    remaining_size: float = betterproto.double_field(6)
    created_at: datetime = betterproto.message_field(7)
    client_order_i_d: str = betterproto.string_field(8)
    open_order_account: str = betterproto.string_field(9)


@dataclass
class GetOrderStatusStreamRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    owner_address: str = betterproto.string_field(2)


@dataclass
class GetOrderStatusStreamResponse(betterproto.Message):
    slot: int = betterproto.int64_field(1)
    order_info: "GetOrderStatusResponse" = betterproto.message_field(2)


@dataclass
class GetOrderStatusResponse(betterproto.Message):
    market: str = betterproto.string_field(1)
    open_order_address: str = betterproto.string_field(2)
    order_i_d: str = betterproto.string_field(3)
    client_order_i_d: int = betterproto.uint64_field(4)
    quantity_released: float = betterproto.float_field(5)
    price: float = betterproto.float_field(6)
    side: "Side" = betterproto.enum_field(7)
    order_status: "OrderStatus" = betterproto.enum_field(8)


@dataclass
class PostSubmitRequest(betterproto.Message):
    transaction: str = betterproto.string_field(1)
    skip_pre_flight: bool = betterproto.bool_field(2)


@dataclass
class PostSubmitResponse(betterproto.Message):
    signature: str = betterproto.string_field(1)


@dataclass
class GetOpenOrdersRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    limit: int = betterproto.uint32_field(2)
    address: str = betterproto.string_field(3)
    open_orders_address: str = betterproto.string_field(4)


@dataclass
class GetOpenOrdersResponse(betterproto.Message):
    orders: List["Order"] = betterproto.message_field(1)


@dataclass
class GetOrderByIDRequest(betterproto.Message):
    order_i_d: str = betterproto.string_field(1)
    market: str = betterproto.string_field(2)


@dataclass
class GetOrderByIDResponse(betterproto.Message):
    order: "Order" = betterproto.message_field(1)


@dataclass
class GetUnsettledRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    owner: str = betterproto.string_field(2)


@dataclass
class UnsettledAccountToken(betterproto.Message):
    address: str = betterproto.string_field(1)
    amount: float = betterproto.double_field(2)


@dataclass
class UnsettledAccount(betterproto.Message):
    account: str = betterproto.string_field(1)
    base_token: "UnsettledAccountToken" = betterproto.message_field(2)
    quote_token: "UnsettledAccountToken" = betterproto.message_field(3)


@dataclass
class GetUnsettledResponse(betterproto.Message):
    market: str = betterproto.string_field(1)
    unsettled: List["UnsettledAccount"] = betterproto.message_field(2)


@dataclass
class GetOrderbooksStreamResponse(betterproto.Message):
    slot: int = betterproto.int64_field(1)
    orderbook: "GetOrderbookResponse" = betterproto.message_field(2)


@dataclass
class GetTickersStreamResponse(betterproto.Message):
    slot: int = betterproto.int64_field(1)
    ticker: "GetTickersResponse" = betterproto.message_field(2)


@dataclass
class GetMarketDepthRequest(betterproto.Message):
    market: str = betterproto.string_field(1)
    depth: int = betterproto.int32_field(2)
    step: "Step" = betterproto.enum_field(3)


@dataclass
class GetMarketDepthStreamResponse(betterproto.Message):
    slot: int = betterproto.int64_field(1)
    tick: "MarketDepthTick" = betterproto.message_field(2)


@dataclass
class MarketDepthTick(betterproto.Message):
    prev_slot: int = betterproto.int64_field(1)
    asks: List["OrderbookItem"] = betterproto.message_field(2)
    bids: List["OrderbookItem"] = betterproto.message_field(3)


@dataclass
class GetTradesStreamResponse(betterproto.Message):
    slot: int = betterproto.int64_field(1)
    trades: "GetTradesResponse" = betterproto.message_field(2)


class ApiStub(betterproto.ServiceStub):
    async def get_markets(self) -> GetMarketsResponse:
        request = GetMarketsRequest()

        return await self._unary_unary(
            "/api.Api/GetMarkets",
            request,
            GetMarketsResponse,
        )

    async def get_tickers(self, *, market: str = "") -> GetTickersResponse:
        request = GetTickersRequest()
        request.market = market

        return await self._unary_unary(
            "/api.Api/GetTickers",
            request,
            GetTickersResponse,
        )

    async def get_kline(
        self,
        *,
        market: str = "",
        from_: Optional[datetime] = None,
        to: Optional[datetime] = None,
        resolution: str = "",
        limit: int = 0,
    ) -> GetKlineResponse:
        request = GetKlineRequest()
        request.market = market
        if from_ is not None:
            request.from_ = from_
        if to is not None:
            request.to = to
        request.resolution = resolution
        request.limit = limit

        return await self._unary_unary(
            "/api.Api/GetKline",
            request,
            GetKlineResponse,
        )

    async def get_orderbook(
        self, *, market: str = "", limit: int = 0
    ) -> GetOrderbookResponse:
        request = GetOrderbookRequest()
        request.market = market
        request.limit = limit

        return await self._unary_unary(
            "/api.Api/GetOrderbook",
            request,
            GetOrderbookResponse,
        )

    async def get_trades(
        self, *, market: str = "", limit: int = 0
    ) -> GetTradesResponse:
        request = GetTradesRequest()
        request.market = market
        request.limit = limit

        return await self._unary_unary(
            "/api.Api/GetTrades",
            request,
            GetTradesResponse,
        )

    async def get_server_time(self) -> GetServerTimeResponse:
        request = GetServerTimeRequest()

        return await self._unary_unary(
            "/api.Api/GetServerTime",
            request,
            GetServerTimeResponse,
        )

    async def get_account_balance(
        self, *, owner_address: str = ""
    ) -> GetAccountBalanceResponse:
        """account endpoints"""

        request = GetAccountBalanceRequest()
        request.owner_address = owner_address

        return await self._unary_unary(
            "/api.Api/GetAccountBalance",
            request,
            GetAccountBalanceResponse,
        )

    async def post_order(
        self,
        *,
        owner_address: str = "",
        payer_address: str = "",
        market: str = "",
        side: "Side" = 0,
        type: List["OrderType"] = [],
        amount: float = 0,
        price: float = 0,
        open_orders_address: str = "",
        client_order_i_d: int = 0,
    ) -> PostOrderResponse:
        """trade endpoints"""

        request = PostOrderRequest()
        request.owner_address = owner_address
        request.payer_address = payer_address
        request.market = market
        request.side = side
        request.type = type
        request.amount = amount
        request.price = price
        request.open_orders_address = open_orders_address
        request.client_order_i_d = client_order_i_d

        return await self._unary_unary(
            "/api.Api/PostOrder",
            request,
            PostOrderResponse,
        )

    async def post_submit(
        self, *, transaction: str = "", skip_pre_flight: bool = False
    ) -> PostSubmitResponse:
        request = PostSubmitRequest()
        request.transaction = transaction
        request.skip_pre_flight = skip_pre_flight

        return await self._unary_unary(
            "/api.Api/PostSubmit",
            request,
            PostSubmitResponse,
        )

    async def post_cancel_order(
        self,
        *,
        order_i_d: str = "",
        side: "Side" = 0,
        market_address: str = "",
        owner_address: str = "",
        open_orders_address: str = "",
    ) -> PostCancelOrderResponse:
        request = PostCancelOrderRequest()
        request.order_i_d = order_i_d
        request.side = side
        request.market_address = market_address
        request.owner_address = owner_address
        request.open_orders_address = open_orders_address

        return await self._unary_unary(
            "/api.Api/PostCancelOrder",
            request,
            PostCancelOrderResponse,
        )

    async def post_cancel_by_client_order_i_d(
        self,
        *,
        client_order_i_d: int = 0,
        market_address: str = "",
        owner_address: str = "",
        open_orders_address: str = "",
    ) -> PostCancelOrderResponse:
        request = PostCancelByClientOrderIDRequest()
        request.client_order_i_d = client_order_i_d
        request.market_address = market_address
        request.owner_address = owner_address
        request.open_orders_address = open_orders_address

        return await self._unary_unary(
            "/api.Api/PostCancelByClientOrderID",
            request,
            PostCancelOrderResponse,
        )

    async def post_cancel_all(
        self,
        *,
        market: str = "",
        owner_address: str = "",
        open_orders_addresses: List[str] = [],
    ) -> PostCancelAllResponse:
        request = PostCancelAllRequest()
        request.market = market
        request.owner_address = owner_address
        request.open_orders_addresses = open_orders_addresses

        return await self._unary_unary(
            "/api.Api/PostCancelAll",
            request,
            PostCancelAllResponse,
        )

    async def post_replace_by_client_order_i_d(
        self,
        *,
        owner_address: str = "",
        payer_address: str = "",
        market: str = "",
        side: "Side" = 0,
        type: List["OrderType"] = [],
        amount: float = 0,
        price: float = 0,
        open_orders_address: str = "",
        client_order_i_d: int = 0,
    ) -> PostOrderResponse:
        request = PostOrderRequest()
        request.owner_address = owner_address
        request.payer_address = payer_address
        request.market = market
        request.side = side
        request.type = type
        request.amount = amount
        request.price = price
        request.open_orders_address = open_orders_address
        request.client_order_i_d = client_order_i_d

        return await self._unary_unary(
            "/api.Api/PostReplaceByClientOrderID",
            request,
            PostOrderResponse,
        )

    async def post_replace_order(
        self,
        *,
        owner_address: str = "",
        payer_address: str = "",
        market: str = "",
        side: "Side" = 0,
        type: List["OrderType"] = [],
        amount: float = 0,
        price: float = 0,
        open_orders_address: str = "",
        client_order_i_d: int = 0,
        order_i_d: str = "",
    ) -> PostOrderResponse:
        request = PostReplaceOrderRequest()
        request.owner_address = owner_address
        request.payer_address = payer_address
        request.market = market
        request.side = side
        request.type = type
        request.amount = amount
        request.price = price
        request.open_orders_address = open_orders_address
        request.client_order_i_d = client_order_i_d
        request.order_i_d = order_i_d

        return await self._unary_unary(
            "/api.Api/PostReplaceOrder",
            request,
            PostOrderResponse,
        )

    async def post_settle(
        self,
        *,
        owner_address: str = "",
        market: str = "",
        base_token_wallet: str = "",
        quote_token_wallet: str = "",
        open_orders_address: str = "",
    ) -> PostSettleResponse:
        request = PostSettleRequest()
        request.owner_address = owner_address
        request.market = market
        request.base_token_wallet = base_token_wallet
        request.quote_token_wallet = quote_token_wallet
        request.open_orders_address = open_orders_address

        return await self._unary_unary(
            "/api.Api/PostSettle",
            request,
            PostSettleResponse,
        )

    async def get_orders(
        self,
        *,
        market: str = "",
        status: "OrderStatus" = 0,
        side: "Side" = 0,
        types: List["OrderType"] = [],
        from_: Optional[datetime] = None,
        limit: int = 0,
        direction: "Direction" = 0,
        address: str = "",
        open_orders_address: str = "",
    ) -> GetOrdersResponse:
        request = GetOrdersRequest()
        request.market = market
        request.status = status
        request.side = side
        request.types = types
        if from_ is not None:
            request.from_ = from_
        request.limit = limit
        request.direction = direction
        request.address = address
        request.open_orders_address = open_orders_address

        return await self._unary_unary(
            "/api.Api/GetOrders",
            request,
            GetOrdersResponse,
        )

    async def get_open_orders(
        self,
        *,
        market: str = "",
        limit: int = 0,
        address: str = "",
        open_orders_address: str = "",
    ) -> GetOpenOrdersResponse:
        request = GetOpenOrdersRequest()
        request.market = market
        request.limit = limit
        request.address = address
        request.open_orders_address = open_orders_address

        return await self._unary_unary(
            "/api.Api/GetOpenOrders",
            request,
            GetOpenOrdersResponse,
        )

    async def get_order_by_i_d(
        self, *, order_i_d: str = "", market: str = ""
    ) -> GetOrderByIDResponse:
        request = GetOrderByIDRequest()
        request.order_i_d = order_i_d
        request.market = market

        return await self._unary_unary(
            "/api.Api/GetOrderByID",
            request,
            GetOrderByIDResponse,
        )

    async def get_unsettled(
        self, *, market: str = "", owner: str = ""
    ) -> GetUnsettledResponse:
        request = GetUnsettledRequest()
        request.market = market
        request.owner = owner

        return await self._unary_unary(
            "/api.Api/GetUnsettled",
            request,
            GetUnsettledResponse,
        )

    async def get_orderbooks_stream(
        self, *, markets: List[str] = [], limit: int = 0
    ) -> AsyncGenerator[GetOrderbooksStreamResponse, None]:
        """streaming endpoints"""

        request = GetOrderbooksRequest()
        request.markets = markets
        request.limit = limit

        async for response in self._unary_stream(
            "/api.Api/GetOrderbooksStream",
            request,
            GetOrderbooksStreamResponse,
        ):
            yield response

    async def get_tickers_stream(
        self, *, market: str = ""
    ) -> AsyncGenerator[GetTickersStreamResponse, None]:
        request = GetTickersRequest()
        request.market = market

        async for response in self._unary_stream(
            "/api.Api/GetTickersStream",
            request,
            GetTickersStreamResponse,
        ):
            yield response

    async def get_market_depth_stream(
        self,
    ) -> AsyncGenerator[GetMarketDepthStreamResponse, None]:
        request = GetMarketsRequest()

        async for response in self._unary_stream(
            "/api.Api/GetMarketDepthStream",
            request,
            GetMarketDepthStreamResponse,
        ):
            yield response

    async def get_trades_stream(
        self, *, market: str = "", limit: int = 0
    ) -> AsyncGenerator[GetTradesStreamResponse, None]:
        request = GetTradesRequest()
        request.market = market
        request.limit = limit

        async for response in self._unary_stream(
            "/api.Api/GetTradesStream",
            request,
            GetTradesStreamResponse,
        ):
            yield response

    async def get_order_status_stream(
        self, *, market: str = "", owner_address: str = ""
    ) -> AsyncGenerator[GetOrderStatusStreamResponse, None]:
        request = GetOrderStatusStreamRequest()
        request.market = market
        request.owner_address = owner_address

        async for response in self._unary_stream(
            "/api.Api/GetOrderStatusStream",
            request,
            GetOrderStatusStreamResponse,
        ):
            yield response



import base64
from typing import List

from solana import transaction as solana_transaction, keypair
from solana.utils import shortvec_encoding as shortvec


class PartialTransaction:
    signatures: List[bytes]
    message: bytes

    def __init__(self, message_bytes: bytes, signatures: List[bytes]):
        self.message = message_bytes
        self.signatures = signatures

    def complete_signing(self, private_key: keypair.Keypair):
        new_signature = private_key.sign(self.message)

        for i, existing_signature in enumerate(self.signatures):
            if is_zero(existing_signature):
                self.signatures[i] = new_signature.signature
                break

    def serialize(self) -> bytes:
        output = bytearray()

        signature_count = shortvec.encode_length(len(self.signatures))
        output.extend(signature_count)

        for signature in self.signatures:
            output.extend(signature)

        output.extend(self.message)
        return bytes(output)

    @classmethod
    def deserialize(cls, tx_base64: str) -> "PartialTransaction":
        tx_bytes = bytes(tx_base64, encoding="utf-8")
        tx_bytes_base64 = base64.decodebytes(tx_bytes)

        signatures = []
        signature_count, offset = shortvec.decode_length(tx_bytes_base64)
        zero_signatures = 0
        for _ in range(signature_count):
            signature_bytes = tx_bytes_base64[
                offset : offset + solana_transaction.SIG_LENGTH
            ]
            if is_zero(signature_bytes):
                zero_signatures += 1

            signatures.append(signature_bytes)
            offset += solana_transaction.SIG_LENGTH

        message_bytes = tx_bytes_base64[offset:]
        num_required_signatures = message_bytes[0]

        if zero_signatures != 1:
            raise ValueError()

        if num_required_signatures != len(signatures):
            raise ValueError()

        return PartialTransaction(message_bytes, signatures)


def is_zero(signature_bytes: bytes) -> bool:
    return signature_bytes == bytes(64)



import os
from typing import TYPE_CHECKING, Optional

from grpclib import client
from solana import keypair

from .. import transaction
from . import constants
from .base import Provider

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from betterproto import _MetadataLike, Deadline


class GrpcProvider(Provider):
    # pyre-ignore[15]: overriding to force context manager hooks
    channel: Optional[client.Channel] = None

    _host: str
    _port: int
    _private_key: Optional[keypair.Keypair]

    def __init__(
        self,
        host: str = constants.MAINNET_API_GRPC_HOST,
        port: int = constants.MAINNET_API_GRPC_PORT,
        private_key: Optional[str] = None,
        auth_header: Optional[str] = None,
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["_MetadataLike"] = None,
    ):
        self._host = host
        self._port = port
        self._auth_header = auth_header

        if private_key is None:
            try:
                self._private_key = transaction.load_private_key_from_env()
            except EnvironmentError:
                self._private_key = None
        else:
            self._private_key = transaction.load_private_key(private_key)

        if auth_header is None:
            self._auth_header = os.environ["AUTH_HEADER"]
        else:
            self._auth_header = auth_header

        super().__init__(
            # pyre-ignore[6]: overriding to force context manager hooks
            None,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def connect(self):
        if self.channel is None:
            self.channel = client.Channel(self._host, self._port)
            self.metadata = {"authorization": self._auth_header}

    def private_key(self) -> Optional[keypair.Keypair]:
        return self._private_key

    async def close(self):
        channel = self.channel
        if channel is not None:
            self.channel.close()


def grpc(auth_header: Optional[str] = None) -> Provider:
    return GrpcProvider(auth_header=auth_header)


def grpc_testnet(auth_header: Optional[str] = None) -> Provider:
    return GrpcProvider(
        host=constants.TESTNET_API_GRPC_HOST,
        port=constants.TESTNET_API_GRPC_PORT,
        auth_header=auth_header,
    )


def grpc_devnet(auth_header: Optional[str] = None) -> Provider:
    return GrpcProvider(
        host=constants.DEVNET_API_GRPC_HOST,
        port=constants.DEVNET_API_GRPC_PORT,
        auth_header=auth_header,
    )


def grpc_local(auth_header: Optional[str] = None) -> Provider:
    return GrpcProvider(
        host=constants.LOCAL_API_GRPC_HOST,
        port=constants.LOCAL_API_GRPC_PORT,
        auth_header=auth_header,
    )



from bxserum import provider

Provider = provider.Provider


async def serum(connection_provider: Provider) -> Provider:
    await connection_provider.connect()
    return connection_provider


_all = ["serum", "Provider"]



MAINNET_API_HTTP = "https://virginia.solana.dex.blxrbdn.com"
MAINNET_API_WS = "wss://virginia.solana.dex.blxrbdn.com/ws"
MAINNET_API_GRPC_HOST = "virginia.solana.dex.blxrbdn.com"
MAINNET_API_GRPC_PORT = 443

TESTNET_API_HTTP = (
    "http://serum-nlb-5a2c3912804344a3.elb.us-east-1.amazonaws.com/"
)
TESTNET_API_WS = (
    "ws://serum-nlb-5a2c3912804344a3.elb.us-east-1.amazonaws.com/ws"
)
TESTNET_API_GRPC_HOST = "serum-nlb-5a2c3912804344a3.elb.us-east-1.amazonaws.com"
TESTNET_API_GRPC_PORT = 80

DEVNET_API_HTTP = (
    "http://serum-nlb-53baf45ef9775263.elb.us-east-1.amazonaws.com/"
)
DEVNET_API_WS = "ws://serum-nlb-53baf45ef9775263.elb.us-east-1.amazonaws.com/ws"
DEVNET_API_GRPC_HOST = "serum-nlb-53baf45ef9775263.elb.us-east-1.amazonaws.com"
DEVNET_API_GRPC_PORT = 80

LOCAL_API_HTTP = "http://127.0.0.1:9000"
LOCAL_API_WS = "ws://127.0.0.1:9000/ws"
LOCAL_API_GRPC_HOST = "127.0.0.1"
LOCAL_API_GRPC_PORT = 9000



from typing import List, Dict, Any

import aiohttp
import betterproto


class HttpError(Exception):
    code: int
    message: str
    details: List[str]

    def __init__(self, code: int, message: str, details: List[str]):
        super().__init__()

        self.code = code
        self.message = message
        self.details = details

    def __str__(self):
        return f"HttpError[{self.code}]: {self.message} ({self.details})"

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "HttpError":
        return cls(
            payload["code"],
            payload["message"],
            payload["details"],
        )


async def map_response(
    response: aiohttp.ClientResponse, destination: betterproto.Message
):
    response_json = await response.json()
    try:
        http_error = HttpError.from_json(response_json)
        raise http_error
    except KeyError:
        return destination.from_dict(response_json)



from .signing import (
    load_private_key,
    load_private_key_from_env,
    sign_tx,
    sign_tx_with_private_key,
    load_open_orders,
)

__all__ = [
    "load_private_key",
    "load_private_key_from_env",
    "sign_tx",
    "sign_tx_with_private_key",
    "load_open_orders",
]



from abc import ABC, abstractmethod
from typing import List, Optional

from solana import keypair

from .. import proto, transaction


class Provider(proto.ApiStub, ABC):
    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, *exc_info):
        await self.close()

    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    def private_key(self) -> Optional[keypair.Keypair]:
        pass

    @abstractmethod
    async def close(self):
        pass

    def require_private_key(self) -> keypair.Keypair:
        kp = self.private_key()
        if kp is None:
            raise EnvironmentError("private key has not been set in provider")
        return kp

    async def submit_order(
        self,
        owner_address: str,
        payer_address: str,
        market: str,
        side: "proto.Side",
        types: List["proto.OrderType"],
        amount: float,
        price: float,
        open_orders_address: str = "",
        client_order_id: int = 0,
        skip_pre_flight: bool = False,
    ) -> str:
        pk = self.require_private_key()
        order = await self.post_order(
            owner_address=owner_address,
            payer_address=payer_address,
            market=market,
            side=side,
            type=types,
            amount=amount,
            price=price,
            open_orders_address=open_orders_address,
            client_order_i_d=client_order_id,
        )
        signed_tx = transaction.sign_tx_with_private_key(order.transaction, pk)
        result = await self.post_submit(
            transaction=signed_tx, skip_pre_flight=skip_pre_flight
        )
        return result.signature

    async def submit_cancel_order(
        self,
        order_i_d: str = "",
        side: proto.Side = proto.Side.S_UNKNOWN,
        market_address: str = "",
        owner_address: str = "",
        open_orders_address: str = "",
        skip_pre_flight: bool = True,
    ) -> str:
        pk = self.require_private_key()
        order = await self.post_cancel_order(
            order_i_d=order_i_d,
            side=side,
            market_address=market_address,
            owner_address=owner_address,
            open_orders_address=open_orders_address,
        )
        signed_tx = transaction.sign_tx_with_private_key(order.transaction, pk)
        result = await self.post_submit(
            transaction=signed_tx, skip_pre_flight=skip_pre_flight
        )
        return result.signature

    async def submit_cancel_by_client_order_i_d(
        self,
        client_order_i_d: int = 0,
        market_address: str = "",
        owner_address: str = "",
        open_orders_address: str = "",
        skip_pre_flight: bool = True,
    ) -> str:
        pk = self.require_private_key()
        order = await self.post_cancel_by_client_order_i_d(
            client_order_i_d=client_order_i_d,
            market_address=market_address,
            owner_address=owner_address,
            open_orders_address=open_orders_address,
        )
        signed_tx = transaction.sign_tx_with_private_key(order.transaction, pk)
        result = await self.post_submit(
            transaction=signed_tx, skip_pre_flight=skip_pre_flight
        )
        return result.signature

    async def submit_cancel_all(
        self,
        market: str = "",
        owner_address: str = "",
        open_orders_addresses: Optional[List[str]] = None,
        skip_pre_flight: bool = True,
    ) -> List[str]:
        if open_orders_addresses is None:
            open_orders_addresses = []

        pk = self.require_private_key()
        response = await self.post_cancel_all(
            market=market,
            owner_address=owner_address,
            open_orders_addresses=open_orders_addresses,
        )

        signatures = []
        for tx in response.transactions:
            signed_tx = transaction.sign_tx_with_private_key(tx, pk)
            result = await self.post_submit(
                transaction=signed_tx, skip_pre_flight=skip_pre_flight
            )
            signatures.append(result.signature)

        return signatures

    async def submit_settle(
        self,
        owner_address: str = "",
        market: str = "",
        base_token_wallet: str = "",
        quote_token_wallet: str = "",
        open_orders_address: str = "",
        skip_pre_flight: bool = False,
    ) -> str:
        pk = self.require_private_key()
        response = await self.post_settle(
            owner_address=owner_address,
            market=market,
            base_token_wallet=base_token_wallet,
            quote_token_wallet=quote_token_wallet,
            open_orders_address=open_orders_address,
        )
        signed_tx = transaction.sign_tx_with_private_key(
            response.transaction, pk
        )
        result = await self.post_submit(
            transaction=signed_tx, skip_pre_flight=skip_pre_flight
        )
        return result.signature

    async def submit_replace_by_client_order_i_d(
        self,
        owner_address: str,
        payer_address: str,
        market: str,
        side: "proto.Side",
        types: List["proto.OrderType"],
        amount: float,
        price: float,
        open_orders_address: str = "",
        client_order_id: int = 0,
        skip_pre_flight: bool = False,
    ) -> str:
        pk = self.require_private_key()
        order = await self.post_replace_by_client_order_i_d(
            owner_address=owner_address,
            payer_address=payer_address,
            market=market,
            side=side,
            type=types,
            amount=amount,
            price=price,
            open_orders_address=open_orders_address,
            client_order_i_d=client_order_id,
        )
        signed_tx = transaction.sign_tx_with_private_key(order.transaction, pk)
        result = await self.post_submit(
            transaction=signed_tx, skip_pre_flight=skip_pre_flight
        )
        return result.signature

    async def submit_replace_order(
        self,
        order_id: str,
        owner_address: str,
        payer_address: str,
        market: str,
        side: "proto.Side",
        types: List["proto.OrderType"],
        amount: float,
        price: float,
        open_orders_address: str = "",
        client_order_id: int = 0,
        skip_pre_flight: bool = False,
    ) -> str:
        pk = self.require_private_key()
        order = await self.post_replace_order(
            owner_address=owner_address,
            payer_address=payer_address,
            market=market,
            side=side,
            type=types,
            amount=amount,
            price=price,
            open_orders_address=open_orders_address,
            client_order_i_d=client_order_id,
            order_i_d=order_id,
        )
        signed_tx = transaction.sign_tx_with_private_key(order.transaction, pk)
        result = await self.post_submit(
            transaction=signed_tx, skip_pre_flight=skip_pre_flight
        )
        return result.signature


class NotConnectedException(Exception):
    pass



import os
import base58
import base64

from solana import keypair as kp

from .deserializer import PartialTransaction


def load_private_key(pkey_str: str) -> kp.Keypair:
    # convert base58 private key string to a keypair
    pkey_bytes = bytes(pkey_str, encoding="utf-8")
    pkey_bytes_base58 = base58.b58decode(pkey_bytes)
    return kp.Keypair.from_secret_key(pkey_bytes_base58)


def load_private_key_from_env() -> kp.Keypair:
    # get base58 encoded private key
    pkey_str = os.getenv("PRIVATE_KEY")
    if pkey_str is None:
        raise EnvironmentError("env variable `PRIVATE_KEY` not set")

    return load_private_key(pkey_str)


def load_open_orders() -> str:
    open_orders = os.getenv("OPEN_ORDERS")
    if open_orders is None:
        raise EnvironmentError("env variable `OPEN_ORDERS` not set")

    return open_orders


def sign_tx(unsigned_tx_base64: str) -> str:
    """
    Uses environment variable `PRIVATE_KEY` to sign message content and replace zero signatures.

    :param unsigned_tx_base64: transaction bytes in base64
    :return: signed transaction
    """
    keypair = load_private_key_from_env()
    return sign_tx_with_private_key(unsigned_tx_base64, keypair)


def sign_tx_with_private_key(
    unsigned_tx_base64: str, keypair: kp.Keypair
) -> str:
    """
    Signs message content and replaces placeholder zero signature with signature.

    :param unsigned_tx_base64: transaction bytes in base64
    :param keypair: key pair to sign with
    :return: signed transaction
    """
    partial_tx = PartialTransaction.deserialize(unsigned_tx_base64)
    partial_tx.complete_signing(keypair)

    # convert transaction back to base64
    signed_tx_bytes_base64 = base64.b64encode(partial_tx.serialize())
    return signed_tx_bytes_base64.decode("utf-8")



import datetime
import os
from typing import Type, AsyncGenerator, Optional, TYPE_CHECKING, List

import aiohttp
from solana import keypair

from .. import proto, transaction
from . import constants
from .base import Provider
from .http_error import map_response

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences,PyProtectedMember
    # pyre-ignore[21]: module is too hard to find
    from grpclib._protocols import IProtoMessage

    # noinspection PyProtectedMember
    from betterproto import _MetadataLike, Deadline, T


class HttpProvider(Provider):
    _endpoint: str
    _session: aiohttp.ClientSession
    _private_key: Optional[keypair.Keypair]

    # noinspection PyMissingConstructor
    def __init__(
        self,
        endpoint: str = constants.MAINNET_API_HTTP,
        auth_header: Optional[str] = None,
        private_key: Optional[str] = None,
    ):
        self._endpoint = f"{endpoint}/api/v1"

        if auth_header is None:
            auth_header = os.environ["AUTH_HEADER"]

        self._session = aiohttp.ClientSession()
        self._session.headers["authorization"] = auth_header

        if private_key is None:
            try:
                self._private_key = transaction.load_private_key_from_env()
            except EnvironmentError:
                self._private_key = None
        else:
            self._private_key = transaction.load_private_key(private_key)

    async def connect(self):
        pass

    def private_key(self) -> Optional[keypair.Keypair]:
        return self._private_key

    async def close(self):
        await self._session.close()

    async def get_markets(self) -> proto.GetMarketsResponse:
        async with self._session.get(f"{self._endpoint}/market/markets") as res:
            return await map_response(res, proto.GetMarketsResponse())

    async def get_orderbook(
        self, *, market: str = "", limit: int = 0
    ) -> proto.GetOrderbookResponse:
        async with self._session.get(
            f"{self._endpoint}/market/orderbooks/{market}?limit={limit}"
        ) as res:
            return await map_response(res, proto.GetOrderbookResponse())

    async def get_tickers(
        self, *, market: str = ""
    ) -> proto.GetTickersResponse:
        async with self._session.get(
            f"{self._endpoint}/market/tickers/{market}"
        ) as res:
            return await map_response(res, proto.GetTickersResponse())

    async def get_orders(
        self,
        *,
        market: str = "",
        status: proto.OrderStatus = proto.OrderStatus.OS_UNKNOWN,
        side: proto.Side = proto.Side.S_UNKNOWN,
        types: List[proto.OrderType] = [],
        from_: Optional[datetime.datetime] = None,
        limit: int = 0,
        direction: proto.Direction = proto.Direction.D_ASCENDING,
        address: str = "",
        open_orders_address: str = "",
    ) -> proto.GetOrdersResponse:
        raise NotImplementedError()

    async def get_open_orders(
        self,
        *,
        market: str = "",
        side: proto.Side = proto.Side.S_UNKNOWN,
        types: List[proto.OrderType] = [],
        from_: Optional[datetime.datetime] = None,
        limit: int = 0,
        direction: proto.Direction = proto.Direction.D_ASCENDING,
        address: str = "",
        open_orders_address: str = "",
    ) -> proto.GetOpenOrdersResponse:
        async with self._session.get(
            f"{self._endpoint}/trade/orders/{market}"
            f"?address={address}"
            f"?openOrdersAddress={open_orders_address}"
            f"&side={side}"
            "&types=OT_LIMIT"
            f"&direction={direction.name}"
        ) as res:
            return await map_response(res, proto.GetOpenOrdersResponse())

    async def get_order_by_i_d(
        self, *, order_i_d: str = "", market: str = ""
    ) -> proto.GetOrderByIDResponse:
        # TODO
        raise NotImplementedError()

    async def get_unsettled(
        self, *, market: str = "", owner: str = ""
    ) -> proto.GetUnsettledResponse:
        async with self._session.get(
            f"{self._endpoint}/trade/unsettled/{market}?owner={owner}"
        ) as res:
            return await map_response(res, proto.GetUnsettledResponse())

    async def get_account_balance(
        self, owner_address: str = ""
    ) -> proto.GetAccountBalanceResponse:
        async with self._session.get(
            f"{self._endpoint}/account/balance?ownerAddress={owner_address}"
        ) as res:
            return await map_response(res, proto.GetAccountBalanceResponse())

    async def post_order(
        self,
        *,
        owner_address: str = "",
        payer_address: str = "",
        market: str = "",
        side: proto.Side = proto.Side.S_UNKNOWN,
        type: List["proto.OrderType"] = [],
        amount: float = 0,
        price: float = 0,
        open_orders_address: str = "",
        client_order_i_d: int = 0,
    ) -> proto.PostOrderResponse:
        request = proto.PostOrderRequest(
            owner_address,
            payer_address,
            market,
            side,
            type,
            amount,
            price,
            open_orders_address,
            client_order_i_d,
        )

        async with self._session.post(
            f"{self._endpoint}/trade/place", json=request.to_dict()
        ) as res:
            return await map_response(res, proto.PostOrderResponse())

    async def post_cancel_order(
        self,
        *,
        order_i_d: str = "",
        side: proto.Side = proto.Side.S_UNKNOWN,
        market_address: str = "",
        owner_address: str = "",
        open_orders_address: str = "",
    ) -> proto.PostCancelOrderResponse:
        request = proto.PostCancelOrderRequest(
            order_i_d,
            side,
            market_address,
            owner_address,
            open_orders_address,
        )

        async with self._session.post(
            f"{self._endpoint}/trade/cancel", json=request.to_dict()
        ) as res:
            return await map_response(res, proto.PostCancelOrderResponse())

    async def post_cancel_by_client_order_i_d(
        self,
        *,
        client_order_i_d: int = 0,
        market_address: str = "",
        owner_address: str = "",
        open_orders_address: str = "",
    ) -> proto.PostCancelOrderResponse:
        request = proto.PostCancelByClientOrderIDRequest(
            client_order_i_d,
            market_address,
            owner_address,
            open_orders_address,
        )

        async with self._session.post(
            f"{self._endpoint}/trade/cancelbyid", json=request.to_dict()
        ) as res:
            return await map_response(res, proto.PostCancelOrderResponse())

    async def post_cancel_all(
        self,
        *,
        market: str = "",
        owner_address: str = "",
        open_orders_addresses: List[str] = [],
    ) -> proto.PostCancelAllResponse:
        request = proto.PostCancelAllRequest(
            market, owner_address, open_orders_addresses
        )
        async with self._session.post(
            f"{self._endpoint}/trade/cancelall", json=request.to_dict()
        ) as res:
            return await map_response(res, proto.PostCancelAllResponse())

    async def post_settle(
        self,
        *,
        owner_address: str = "",
        market: str = "",
        base_token_wallet: str = "",
        quote_token_wallet: str = "",
        open_orders_address: str = "",
    ) -> proto.PostSettleResponse:
        request = proto.PostSettleRequest(
            owner_address,
            market,
            base_token_wallet,
            quote_token_wallet,
            open_orders_address,
        )
        async with self._session.post(
            f"{self._endpoint}/trade/settle", json=request.to_dict()
        ) as res:
            return await map_response(res, proto.PostSettleResponse())

    async def post_submit(
        self, *, transaction: str = "", skip_pre_flight: bool = False
    ) -> proto.PostSubmitResponse:
        request = proto.PostSubmitRequest(transaction, skip_pre_flight)
        async with self._session.post(
            f"{self._endpoint}/trade/submit", json=request.to_dict()
        ) as res:
            return await map_response(res, proto.PostSubmitResponse())

    async def post_replace_by_client_order_i_d(
        self,
        *,
        owner_address: str = "",
        payer_address: str = "",
        market: str = "",
        side: proto.Side = proto.Side.S_UNKNOWN,
        type: List["proto.OrderType"] = [],
        amount: float = 0,
        price: float = 0,
        open_orders_address: str = "",
        client_order_i_d: int = 0,
    ) -> proto.PostOrderResponse:
        request = proto.PostOrderRequest(
            owner_address,
            payer_address,
            market,
            side,
            type,
            amount,
            price,
            open_orders_address,
            client_order_i_d,
        )

        async with self._session.post(
            f"{self._endpoint}/trade/replacebyclientid", json=request.to_dict()
        ) as res:
            return await map_response(res, proto.PostOrderResponse())

    async def post_replace_order(
        self,
        *,
        owner_address: str = "",
        payer_address: str = "",
        market: str = "",
        side: proto.Side = proto.Side.S_UNKNOWN,
        type: List["proto.OrderType"] = [],
        amount: float = 0,
        price: float = 0,
        open_orders_address: str = "",
        client_order_i_d: int = 0,
        order_i_d: str,
    ) -> proto.PostOrderResponse:
        request = proto.PostReplaceOrderRequest(
            owner_address,
            payer_address,
            market,
            side,
            type,
            amount,
            price,
            open_orders_address,
            client_order_i_d,
            order_i_d,
        )

        async with self._session.post(
            f"{self._endpoint}/trade/replace", json=request.to_dict()
        ) as res:
            return await map_response(res, proto.PostOrderResponse())

    async def _unary_stream(
        self,
        route: str,
        # pyre-ignore[11]: type is too hard to find
        request: "IProtoMessage",
        response_type: Type["T"],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["_MetadataLike"] = None,
    ) -> AsyncGenerator["T", None]:
        raise NotImplementedError(
            "streaming is not implemented in HTTP provider"
        )

        # useless line to turn function into a generator
        yield response_type()


def http() -> Provider:
    return HttpProvider()


def http_testnet() -> Provider:
    return HttpProvider(endpoint=constants.TESTNET_API_HTTP)


def http_devnet() -> Provider:
    return HttpProvider(endpoint=constants.DEVNET_API_HTTP)


def http_local() -> Provider:
    return HttpProvider(endpoint=constants.LOCAL_API_HTTP)



import os
from typing import TYPE_CHECKING, Type, Optional, AsyncGenerator

import jsonrpc
from solana import keypair

from . import Provider, constants
from .. import transaction

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences,PyProtectedMember
    # pyre-ignore[21]: module is too hard to find
    from grpclib._protocols import IProtoMessage

    # noinspection PyProtectedMember
    from betterproto import _MetadataLike, Deadline, T


class WsProvider(Provider):
    _ws: jsonrpc.WsRpcConnection

    _endpoint: str
    _private_key: Optional[keypair.Keypair]

    # noinspection PyMissingConstructor
    def __init__(
        self,
        endpoint: str = constants.MAINNET_API_WS,
        auth_header: Optional[str] = None,
        private_key: Optional[str] = None,
    ):
        self._endpoint = endpoint

        if auth_header is None:
            auth_header = os.environ["AUTH_HEADER"]

        opts = jsonrpc.WsRpcOpts(headers={"authorization": auth_header})
        self._ws = jsonrpc.WsRpcConnection(endpoint, opts)

        if private_key is None:
            try:
                self._private_key = transaction.load_private_key_from_env()
            except EnvironmentError:
                self._private_key = None
        else:
            self._private_key = transaction.load_private_key(private_key)

    async def connect(self):
        await self._ws.connect()

    def private_key(self) -> Optional[keypair.Keypair]:
        return self._private_key

    async def close(self):
        await self._ws.close()

    async def _unary_unary(
        self,
        route: str,
        # pyre-ignore[11]: type is too hard to find
        request: "IProtoMessage",
        response_type: Type["T"],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["_MetadataLike"] = None,
    ) -> "T":
        result = await self._ws.call(
            _ws_endpoint(route), request.to_dict(include_default_values=False)
        )
        return response_type().from_dict(result)

    async def _unary_stream(
        self,
        route: str,
        request: "IProtoMessage",
        response_type: Type["T"],
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["_MetadataLike"] = None,
    ) -> AsyncGenerator["T", None]:
        subscription_id = await self._ws.subscribe(
            _ws_endpoint(route), request.to_dict()
        )
        async for update in self._ws.notifications_for_id(subscription_id):
            yield response_type().from_dict(update)


def _ws_endpoint(route: str) -> str:
    return route.split("/")[-1]


def ws() -> Provider:
    return WsProvider()


def ws_testnet() -> Provider:
    return WsProvider(endpoint=constants.TESTNET_API_WS)


def ws_devnet() -> Provider:
    return WsProvider(endpoint=constants.DEVNET_API_WS)


def ws_local() -> Provider:
    return WsProvider(endpoint=constants.LOCAL_API_WS)



from .base import Provider
from .grpc import GrpcProvider, grpc, grpc_local, grpc_testnet, grpc_devnet
from .http import HttpProvider, http, http_local, http_testnet, http_devnet
from .http_error import HttpError
from .ws import WsProvider, ws, ws_local, ws_testnet, ws_devnet

__all__ = [
    "Provider",
    "GrpcProvider",
    "grpc",
    "grpc_devnet",
    "grpc_local",
    "grpc_testnet",
    "HttpProvider",
    "HttpError",
    "http",
    "http_devnet",
    "http_local",
    "http_testnet",
    "WsProvider",
    "ws",
    "ws_devnet",
    "ws_local",
    "ws_testnet",
]
